#include "core.h"
#include "SDL/SDL_opengl.h"
#include <math.h>
#include "math_utils.h"
#include "gl_utils.h"
#include "Target.h"

Robot::Robot()
{
	geoms.resize(4);
	left_encoder = right_encoder = 0;
	ticks_per_revolution = 41000;
	llan = lran = 0;
	first_encoder_read = true;
	lm_cmdvel = rm_cmdvel = 0;
	enable_trail = true;
	// 	maxage = 10;
	maxage = 1000;
	history_per_sec = 2;
	//max_wheel_revs_per_sec = 1;
	max_wheel_revs_per_sec = 3;
	rc_m = SDL_CreateMutex();
	history_m = SDL_CreateMutex();
}

Robot::~Robot()
{
}

void Robot::SetupODE(dWorldID world, dSpaceID world_space)
{
	double width = 0.4, length = 0.7, thickness = 0.02;
	double height = 0.4;
	wheel_radius = 0.07;
	axis_length = width;
	dMass m;
	bodies[0] = dBodyCreate(world);
	dBodySetPosition(bodies[0], -0.5*length, 0, height);
	dMassSetBox(&m, 1, length, width, thickness);
	dMassAdjust(&m, 5);
	dBodySetMass(bodies[0], &m);
	geoms[0] = dCreateBox(0, length, width, thickness);
	dGeomSetBody(geoms[0], bodies[0]);

	for(int i=1; i<=3; i++) {
		bodies[i] = dBodyCreate(world);
		dQuaternion q;
		dQFromAxisAndAngle(q, 1, 0, 0, M_PI*0.5);
		dBodySetQuaternion(bodies[i], q);
		dMassSetSphere(&m, 1, wheel_radius);
		dMassAdjust(&m, 0.2);
		dBodySetMass(bodies[i], &m);
		geoms[i] = dCreateSphere(0, wheel_radius);
		dGeomSetBody(geoms[i], bodies[i]);
	}
	
	dBodySetPosition(bodies[1], -length,          0, height);
	dBodySetPosition(bodies[2],  0,  0.5*width, height);
	dBodySetPosition(bodies[3],  0, -0.5*width, height);
	

	for(int i=0; i<3; i++) {
		wheel_joints[i] = dJointCreateHinge(world, 0);
		dJointAttach(wheel_joints[i], bodies[0], bodies[i+1]);
		const dReal *a = dBodyGetPosition(bodies[i+1]);
		dJointSetHingeAnchor(wheel_joints[i], a[0], a[1], a[2]);
		dJointSetHingeAxis(wheel_joints[i], 0, 1, 0);
	}

	for(int i=1; i<3; i++) {
		//dJointSetHinge2Param(wheel_joints[i], dParamLoStop, 0);
		//dJointSetHinge2Param(wheel_joints[i], dParamHiStop, 0);
	}

	robot_space = dSimpleSpaceCreate(world_space);
	dSpaceSetCleanup(robot_space, 0);
	for(int i=0; i<4; i++) {
		dSpaceAdd(robot_space, geoms[i]);
	}
}
	
void Robot::Draw() {
	const dReal *p = dGeomGetPosition(geoms[0]);
	const dReal *r = dGeomGetRotation(geoms[0]);

	setTransform(p, r);
	glColor3f(1, 0, 0);
	glBegin(GL_POLYGON);
	glVertex2d(-0.35, -0.2);
	glVertex2d(-0.35,  0.2);
	glVertex2d( 0.35,  0.2);
	glVertex2d( 0.35, -0.2);
	glEnd();
	glColor3f(0, 1, 0);
	glBegin(GL_LINES);
	glVertex2d(0, 0);
	glVertex2d(0.5, 0);
	glVertex2d(0.5, 0);
	glVertex2d(0.4, -0.1);
	glVertex2d(0.5, 0);
	glVertex2d(0.4, 0.1);
	glEnd();

	glPopMatrix();
	glPushMatrix();

	for(int i=1; i<4; i++) {
		glPopMatrix();
		glPushMatrix();
		const dReal *p = dGeomGetPosition(geoms[i]);
		const dReal *r = dGeomGetRotation(geoms[i]);
		setTransform(p, r);
		glColor3f(i==3?1:0, i==1?1:0, i==2?1:0);
		glBegin(GL_POLYGON);
		glVertex2f(-0.1, -0.1);
		glVertex2f(-0.1,  0.1);
		glVertex2f( 0.1,  0.1);
		glVertex2f( 0.1, -0.1);
		glEnd();
	}
	
	glPopMatrix();
	glPushMatrix();
	if(enable_trail) {
	  
	  GLfloat line_width=5;
	  glLineWidth(line_width);

		CleanupHistory(maxage);
		double now = dtime();
		map<double,pair<double, double> >::iterator it;
		glBegin(GL_LINE_STRIP);
		lock_history();
		for(it = history.begin(); it != history.end(); it++) {
			unlock_history();
			glColor4f(0, 0, 1.0, sqrt(1.0 - ((now - it->first) / maxage)));
			glVertex2f(it->second.first, it->second.second);
			lock_history();
		}
		unlock_history();
		glEnd();
	}
}

void Robot::CleanupHistory(double maxage)
{
	double now = dtime();
	lock_history();
	while(history.size() && history.begin()->first < now - maxage)
		history.erase(history.begin());
	unlock_history();
}

void Robot::SetLeftWheelVel(double rps)
{
	if(rps >  max_wheel_revs_per_sec) rps =  max_wheel_revs_per_sec;
	if(rps < -max_wheel_revs_per_sec) rps = -max_wheel_revs_per_sec;
	dJointSetHingeParam(wheel_joints[1], dParamVel, rps * 2 * M_PI);
	dJointSetHingeParam(wheel_joints[1], dParamFMax, 1);
	lm_cmdvel = rps;
}

void Robot::SetRightWheelVel(double rps)
{
	if(rps >  max_wheel_revs_per_sec) rps =  max_wheel_revs_per_sec;
	if(rps < -max_wheel_revs_per_sec) rps = -max_wheel_revs_per_sec;
	dJointSetHingeParam(wheel_joints[2], dParamVel, -rps * 2 * M_PI);
	dJointSetHingeParam(wheel_joints[2], dParamFMax, 1);
	rm_cmdvel = rps;
}

double Robot::GetLeftActualVel()
{
	return dJointGetHingeAngleRate(wheel_joints[1]) / (2*M_PI);
}

double Robot::GetRightActualVel()
{
	return dJointGetHingeAngleRate(wheel_joints[2]) / (-2*M_PI);
}

void Robot::UpdatePositions()
{
	const dReal *p1, *p2;
	p1 = dGeomGetPosition(geoms[2]);
	p2 = dGeomGetPosition(geoms[3]);

	x = (p1[0] + p2[0])/2.0;
	y = (p1[1] + p2[1])/2.0;
	z = (p1[2] + p2[2])/2.0;

	if(enable_trail) {
		double now = dtime();
		lock_history();
		if(history.size() == 0 || history.rbegin()->first + (1.0/history_per_sec) < now)
			history[now] = make_pair(x, y);
		unlock_history();
	}

	dQuaternion q;
	dGeomGetQuaternion(geoms[0], q);

	double s, ax, ay, az, al;
	s = pow(q[1], 2) + pow(q[2], 2) + pow(q[3], 2);
	ax = q[1] / s;
	ay = q[2] / s;
	az = q[3] / s;
	al = 2 * acos(q[0]);
	
	theta = az > 0 ? 2*M_PI - al: al; // not completely correct, but close enough

	while(theta >  M_PI) theta -= 2*M_PI;
	while(theta < -M_PI) theta += 2*M_PI;
	
	dQuaternion lw, rw;
	dGeomGetQuaternion(geoms[2], lw);
	dGeomGetQuaternion(geoms[3], rw);
	dQuaternion iv, resl, resr;
	double lan, ran;

	iv[0] = q[0];
	iv[1] = -q[1];
	iv[2] = -q[2];
	iv[3] = -q[3];

	QuaternionMult(resl, lw, iv);
	QuaternionMult(resr, rw, iv);
	
	lan = atan2(resl[0], resl[3]) * 2;
	while(lan >  M_PI) lan -= 2*M_PI;
	while(lan < -M_PI) lan += 2*M_PI;

	ran = atan2(resr[0], resr[3]) * 2;
	while(ran >  M_PI) ran -= 2*M_PI;
	while(ran < -M_PI) ran += 2*M_PI;

	double ladiff = llan - lan;
	double radiff = lran - ran;

	if(ladiff < -M_PI) ladiff += 2*M_PI;
	if(ladiff >  M_PI) ladiff -= 2*M_PI;
	if(radiff < -M_PI) radiff += 2*M_PI;
	if(radiff >  M_PI) radiff -= 2*M_PI;

	left_encoder += (long)(2 * M_PI * ladiff * ticks_per_revolution);
	right_encoder -= (long)(2 * M_PI * radiff * ticks_per_revolution);
	
	llan = lan;
	lran = ran;

	if(first_encoder_read) {
		first_encoder_read = false;
		left_encoder = right_encoder = 0;
	}
}

void Robot::PrepareForCollision()
{
}

void Robot::PostCollision()
{
}

int Robot::HandleCollision(dGeomID you, int, Object *other, dSurfaceParameters &surface)
{
	if(other->GetType() == OBJECT_GROUND) {
		surface.mode = dContactApprox1;
		if(GetGeomType(you) == ROBOT_GEOM_FREEWHEEL) {
			surface.mu = 0;
		} else if(GetGeomType(you) == ROBOT_GEOM_DRIVEWHEEL) {
			surface.mode |= dContactSlip1 | dContactSlip2;
			surface.slip1 = 0.01;
			surface.slip2 = 0.01;
			surface.mu = 40;
		} else if(GetGeomType(you) == ROBOT_GEOM_BODY) {
			return COLLISION_HANDLED_IGNORE;
		}
		return COLLISION_HANDLED;
	} else if(other->GetType() == OBJECT_ROBOT) {
		return COLLISION_HANDLED_IGNORE;
	} else if(other->GetType() == OBJECT_TARGET) {
		lock_rc_m();
		if(((Target *)other)->IsHit() == false) {
			recent_collisions.push(OBJECT_TARGET);
		}
		unlock_rc_m();
		// don't handle, just interested in the event
	} else if(other->GetType() == OBJECT_WALL) {
		lock_rc_m();
		recent_collisions.push(OBJECT_WALL);
		unlock_rc_m();
		string s = "";
		if(you == geoms[0]) s = "body";
		if(you == geoms[1]) s = "freewheel";
		if(you == geoms[2]) s = "left wheel";
		if(you == geoms[3]) s = "right wheel";

		//fprintf(stderr, "hit %s\n", s.c_str());
		
		surface.mu = 0;
		surface.mode |= dContactSoftERP;
		surface.soft_erp = 0.9;
		surface.soft_cfm = 0.05;
		surface.mode |= dContactSoftCFM | dContactApprox1;

		if(GetGeomType(you) == ROBOT_GEOM_BODY) return COLLISION_HANDLED_IGNORE;
		
		return COLLISION_HANDLED;
	}
	return COLLISION_UNHANDLED;
}

